@@ -19,10 +19,12 @@
 #include <config.h>
 #include <locale.h>
 #include <stdlib.h>
+#include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
 #include <glib/gi18n.h>
+#include <sys/file.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
@@ -373,7 +375,7 @@ static GdkPixbuf * get_background_pixbuf(void)
 {
     /* Get the root window pixmap. */
     GdkScreen * screen = gdk_screen_get_default();
-#ifdef ENABLE_GTK3
+#ifdef USE_GTK3
     GdkPixbuf * pixbuf = gdk_pixbuf_get_from_window(
         gdk_get_default_root_window(),
         0,
@@ -440,6 +442,14 @@ gboolean expose_event(GtkWidget * widget, GdkEventExpose * event, GdkPixbuf * pi
     return FALSE;
 }
 
+static char lockfile[PATH_MAX];
+
+/* Unlink lockfile on exit. */
+static void main_at_exit(void)
+{
+    unlink(lockfile);
+}
+
 /* Main program. */
 int main(int argc, char * argv[])
 {
@@ -470,6 +480,18 @@ int main(int argc, char * argv[])
     const char * p = g_getenv("_LXSESSION_PID");
     if (p != NULL) handler_context.lxsession_pid = atoi(p);
 
+    /* Create lock file to prevent more than one logout dialog per lxsession process. */
+    sprintf(lockfile, "/tmp/.lxsession-logout-%d.lock", handler_context.lxsession_pid);
+    int fd = open(lockfile, O_RDONLY|O_CREAT, 00600);
+    if (fd >= 0)
+    {
+        if (flock(fd, LOCK_EX | LOCK_NB))
+        {
+            exit(EXIT_FAILURE);
+        }
+    }
+    atexit(main_at_exit);
+
     /* Initialize capabilities of the systemd mechanism. */
     if (dbus_systemd_CanPowerOff())
     {
@@ -579,6 +601,7 @@ int main(int argc, char * argv[])
     gtk_window_set_default_size(GTK_WINDOW(window), gdk_screen_get_width(screen), gdk_screen_get_height(screen));
     gtk_widget_set_app_paintable(window, TRUE);
     g_signal_connect(G_OBJECT(window), "expose_event", G_CALLBACK(expose_event), pixbuf);
+    g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(gtk_main_quit), NULL);
 
     /* Toplevel container */
     GtkWidget* alignment = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
@@ -760,6 +783,15 @@ int main(int argc, char * argv[])
     GtkWidget * cancel_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
     gtk_button_set_alignment(GTK_BUTTON(cancel_button), 0.0, 0.5);
     g_signal_connect(G_OBJECT(cancel_button), "clicked", G_CALLBACK(cancel_clicked), NULL);
+    GtkAccelGroup* accel_group = gtk_accel_group_new();
+    gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
+#if GTK_CHECK_VERSION(3,0,0)
+    gtk_widget_add_accelerator(cancel_button, "activate", accel_group,
+        GDK_KEY_Escape, (GdkModifierType)0, GTK_ACCEL_VISIBLE);
+#else
+    gtk_widget_add_accelerator(cancel_button, "activate", accel_group,
+        GDK_Escape, (GdkModifierType)NULL, GTK_ACCEL_VISIBLE);
+#endif
     gtk_box_pack_start(GTK_BOX(controls), cancel_button, FALSE, FALSE, 4);
 
     /* Create the error text. */
